/*****************************************************************************/
/*FILE: HAL_SYM.C    Routines for the Symbol table                           */
/****************************************************************************
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.
 *****************************************************************************/
/* Originally Created by: Homayoon Akhiani	April,1992		     */
/*****************************************************************************
! REVISION HISTORY:
! Who   When            What
!---------------------------------------------------------------
! HA	16-Jul-1992	HW_LD,HW_ST
! HA	12-May-1992	forward references resolved
! HA    23-Apr-1992	Include stdlib.h (for portability)    
! HA    xx-Apr-1992     First pass
 *****************************************************************************/
/*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "hal.h"

extern FILE		*Ot;		/* Output for Symbol table transaction*/
extern FILE		*Olist;		/* Output for Listing */
extern int		line_number;	/* Source line number */
extern unsigned int	hpc,pc,ic;
extern unsigned int	*m;			/* THE IMAGE */
extern unsigned int	Saddr;
extern avl_tree		*SymTable;		/* AVL data base for symbols*/

#define OBAD	0
#define GOOD	1

int eval();
int symadd();
int symsub();
int symand();
int symorr();
int symxor();
int symone();
/*****************************************************************************/
/* look if the symbol is in, (YES) update the ref, (NO) create  */
/*****************************************************************************/
symuse(s,type)
char *s;    
int  type;	/* type */
{
SENTRY *e;
RNODE  *n, *p;
#if (CASE_SENSITIVE == 0)
 uppercase(s);
#endif		   
 if((avl_lookup(SymTable,s,&e)) == 0)
  { /* Create the symbol entry and add reference */
   symcreate(&e);
   if((n = (RNODE *)malloc(sizeof(RNODE))) == 0)   merror("No memory");
   n->next = NULL;
   n->line = line_number;
   strcpy(n->instr,"NA");
   n->type = type;
   n->ic   = ic;
   n->pref = NULL;	    /* not yet used ???? */
   e->ref  = n;
   strcpy(e->symbol,s);
   avl_insert(SymTable,e->symbol,e); 
   fprintf(Ot,"InsertSymbolUse %s at line %d type=%d\n",
		    e->symbol,line_number,type);
  }
 else
  {	/* entry already exists*/
   fprintf(Ot,"SymbolUse %s at line %d type=%d\n",
		    e->symbol,line_number,type);
   if(e->ref == NULL)
    {
     if((n = (RNODE *)malloc(sizeof(RNODE))) == 0)   merror("No memory");
     n->next = NULL;
     n->line = line_number;
     strcpy(n->instr,"NA");
     n->type = type;
     n->ic   = ic;
     n->pref = NULL;	    /* not yet used ???? */
     e->ref  = n;
    }
   else
    {
     for(p=e->ref;p->next != NULL ;p = p->next);   /*goto the end of the list*/
     if((n = (RNODE *)malloc(sizeof(RNODE))) == 0)   merror("No memory");
     n->next = NULL;
     n->line = line_number;
     strcpy(n->instr,"NA");
     n->type = type;
     n->ic   = ic;
     n->pref = NULL;	    /* not yet used ???? */
     p->next = n;
    };
  };
}
/*****************************************************************************/
/* look if the symbol is in, (YES) check! and update the value, (NO) create  */
/*****************************************************************************/
symdef(s,hv,lv,sl,sr,oper)
char         *s;
unsigned int hv, lv;
char	     *sl,*sr;
int	     oper;
{
SENTRY *e;
#if (CASE_SENSITIVE == 0)
 uppercase(s);
 if(sl!=NULL) uppercase(sl);
 if(sr!=NULL) uppercase(sr);
#endif
 if((avl_lookup(SymTable,s,&e)) == 0)
  {
   symcreate(&e);
   strcpy(e->symbol,s);
   e->pc       = pc;
   e->valid    = 1;
   e->value[0] = lv;
   e->value[1] = hv;
   strcpy(e->s1,"____");
   strcpy(e->s2,"____");
   if(oper != NONE)
    {
     e->operation = oper;
     assert(sl != NULL);
     strcpy(e->s1,sl);
     if(oper != OONE) strcpy(e->s2,sr);
     e->valid     = 0;
     fprintf(Ot,"Defered %s at line %d [%s %s]\n",
		    e->symbol,line_number,sl,sr);

    };
   avl_insert(SymTable,e->symbol,e); 
   fprintf(Ot,"InsertSymbolDef %s at line %d [%08X %08X]\n",
		    e->symbol,line_number,hv,lv);
  }
 else
  {	/* entry already exists, ERROR */
   if(e->valid)
     printf("\n\n\n****Multiple Declarations of %s at lines %d and %d\n",
	e->symbol,line_number,0);
   else
    {
     e->pc       = pc;
     e->valid    = 1;
     e->value[0] = lv;
     e->value[1] = hv;
     strcpy(e->s1,"____");
     strcpy(e->s2,"____");
     if(oper != NONE)
      {
       e->operation = oper;
       assert(sl != NULL);
       strcpy(e->s1,sl);
       if(oper != OONE) strcpy(e->s2,sr);
       e->valid     = 0;
       fprintf(Ot,"Defered %s at line %d [%s %s]\n",
		    e->symbol,line_number,sl,sr);

      };
     fprintf(Ot,"ExistSymbolDef %s at line %d [%08X %08X]\n",
		    e->symbol,line_number,hv,lv);
    };
  };
}
/*****************************************************************************/
/* SYMGET:								     */
/* return 1:	Can't find it, pv is invalid				     */
/* return 2:	Bingo, found it, pv is valid				     */
/*****************************************************************************/
int symget(s,pv)
char         *s;
unsigned int pv[];
{
SENTRY *e;
#if (CASE_SENSITIVE == 0)
 uppercase(s); 
#endif
 if((avl_lookup(SymTable,s,&e)) == 0)
  {
   fprintf(Ot,"UndefineSymbol: %s at line %d\n",s,line_number);
   return(1);
  }
 else
  {	/* entry already exists*/
   if(e->valid)
    {
     pv[0] = e->value[0];
     pv[1] = e->value[1];
    }
   else
    {
     fprintf(Ot,"UndefineSymbol: %s at line %d\n",s,line_number);
     return(1);
    };
  };
 return(2);
}
/*****************************************************************************/
/* SYMRESOLVE:								     */
/*****************************************************************************/
void symresolve(s,e)
char   *s;
SENTRY *e;
{
RNODE *n;
unsigned int offset;
#if (CASE_SENSITIVE == 0)
 uppercase(s);
#endif
 fprintf(Ot,"Resolving %s ->[%08X %08X] %d\n",
		    e->symbol,e->value[1],e->value[0],e->valid);
 switch(e->operation)
  {
   case NONE: break;
   case OADD: e->valid = symadd(e->s1,e->s2,e->value); break;
   case OSUB: e->valid = symsub(e->s1,e->s2,e->value); break;
   case OAND: e->valid = symand(e->s1,e->s2,e->value); break;
   case OORR: e->valid = symorr(e->s1,e->s2,e->value); break;
   case OXOR: e->valid = symxor(e->s1,e->s2,e->value); break;
   case OONE: e->valid = symone(e->s1,e->s2,e->value); break;
   default: printf("HAL_SYM.C: Operation is not supported yet\n"); break;
  };
 if(e->valid == 0) printf("%%Error Undefine Symbol: %s \n",e->symbol);
 for(n=e->ref;n != NULL ;n = n->next)
  {
   switch(n->type)
    {
     case TD32:	
		fprintf(Ot,"D32 at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic]=e->value[0];
		fprintf(Ot,"[%08X]\n",m[n->ic]);
	break;
     case TD64:	
		fprintf(Ot,"D64 at %08X [%08X]->",ic*4,m[n->ic+1],m[n->ic]);
		m[n->ic]=e->value[0]; m[n->ic+1]=e->value[1];	
		fprintf(Ot,"[%08X %08X]\n",m[n->ic+1],m[n->ic]);
	break;
     case TMDS:	/* Memory: Do not clean the m[] (HA)*/
		fprintf(Ot,"MDS at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic] |= (e->value[0] & 0xFFFF);
		fprintf(Ot,"[%08X]\n",m[n->ic]);
	break;
     case TMHW:	/* HW_LD/ST: Do not clean the m[] (HA)*/
		fprintf(Ot,"MHW at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic] |= (e->value[0] & 0xFFF);
		fprintf(Ot,"[%08X]\n",m[n->ic]);
	break;
     case TMOD:	/* HW_LD/ST: Do not clean the m[] (HA)*/
		fprintf(Ot,"MOD at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic] |= (e->value[0] & 0xF)<< 12;
		fprintf(Ot,"[%08X]\n",m[n->ic]);
	break;
     case TBDS:	
		fprintf(Ot,"BDS at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic] &= ~0x1FFFFF; /*clean first */
		offset = (e->value[0] - ((n->ic+1)*4 + Saddr))>>2;
		m[n->ic] |= (offset & 0x1FFFFF);
		fprintf(Ot,"[%08X] offset=%X\n",m[n->ic],offset);
	break;
     case TIOP:	
		fprintf(Ot,"IOP at %08X [%08X]->",ic*4,m[n->ic]);
		m[n->ic] &= 0xFFE01FFF; /*clean first */
		m[n->ic] |= ((e->value[0] & 0xFF) << 13);
		fprintf(Ot,"[%08X]\n",m[n->ic]);
	break;
     default: printf("Bad Tag! symbol %s at line %d \n",
		    e->symbol,n->line); 
	break;
    };  
  }; 
}
/*****************************************************************************/
/*****************************************************************************/
void symtableit(s,e)
char   *s;
SENTRY *e;
{
RNODE *n;
unsigned int offset;

 if(e->valid)
   fprintf(Olist,"%-30s ->[%08X %08X]\n",e->symbol,e->value[1],e->value[0]);
 else
   fprintf(Olist,"%-30s -> No Value! \n",e->symbol);
 for(n=e->ref;n != NULL ;n = n->next)
  {
   switch(n->type)
    {
     case TD32:	fprintf(Olist,"\t\tUsed as 32bit data at (ic) %d\n",n->ic);
	break;
     case TD64:	fprintf(Olist,"\t\tUsed as 64bit data at (ic) %d\n",n->ic);
	break;
     case TMDS:	fprintf(Olist,"\t\tUsed as Displacement at (ic) %d\n",n->ic);
	break;
     case TBDS:	fprintf(Olist,"\t\tUsed as Branch Displacement at (ic) %d\n",
			n->ic);
	break;
     case TIOP:	fprintf(Olist,"\t\tUsed as Immediate data at (ic) %d\n",
		    n->ic);
	break;
     default: 
		fprintf(Olist,"\t\tBad Symbol! at (ic) %d\n",n->ic);
	break;
    };  
  }; 
}
/*****************************************************************************/
symcreate(pe)
SENTRY **pe;
{
 if((*pe = (SENTRY *)malloc(sizeof(SENTRY))) == 0) 
	merror("SYMCREATE: No memory");
 (*pe)->pc	    = 0;
 (*pe)->valid	    = 0;
 (*pe)->value[0]    = 0;
 (*pe)->value[1]    = 0;
 (*pe)->operation   = NONE;
 (*pe)->s1[0]	    = '\0';;
 (*pe)->s2[0]	    = '\0';;
 (*pe)->symbol[0]   = '\0';;
 (*pe)->ref	    = NULL;
}
/*****************************************************************************/
/*****************************************************************************/
/* Symbolics Evaluation Routines:					     */
/*****************************************************************************/
/*****************************************************************************/
/*EVAL:									     */
/*  Evaluate the symbol s, put the 64 bit value in v.			     */
/*****************************************************************************/
int eval(s,v)
char s[];
B32  v[];
{
 if(symget(s,v) != 2)
  {
   if(!strcmp(s,"____")) {printf("Can't evalute"); return(OBAD); };
   scan64(s,v);
  };
 return(GOOD);
}
/*****************************************************************************/
symadd(s1,s2,v) 
char     s1[], s2[];
B32	 v[];
{
B32 s1v[2], s2v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 if(eval(s2,s2v) == OBAD) return(OBAD);
 status = ADD64(s1v,s2v,v);
 if(status != 0) printf("SYMADD: Integer Overflow!\n");
 return(GOOD);
}
/*****************************************************************************/
symsub(s1,s2,v) 
char     s1[], s2[];
B32      v[];
{
B32 s1v[2], s2v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 if(eval(s2,s2v) == OBAD) return(OBAD);
 status = SUB64(s1v,s2v,v);
 if(status != 0) printf("SYMSUB: Integer Overflow!\n");
 return(GOOD);
}
/*****************************************************************************/
symand(s1,s2,v) 
char     s1[], s2[];
B32      v[];
{
B32 s1v[2], s2v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 if(eval(s2,s2v) == OBAD) return(OBAD);
 v[0] = s1v[0] & s2v[0];
 v[1] = s1v[1] & s2v[1];
 return(GOOD);
}
/*****************************************************************************/
symorr(s1,s2,v) 
char     s1[], s2[];
B32      v[];
{
B32 s1v[2], s2v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 if(eval(s2,s2v) == OBAD) return(OBAD);
 v[0] = s1v[0] | s2v[0];
 v[1] = s1v[1] | s2v[1];
 return(GOOD);
}
/*****************************************************************************/
symxor(s1,s2,v) 
char     s1[], s2[];
B32      v[];
{
B32 s1v[2], s2v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 if(eval(s2,s2v) == OBAD) return(OBAD);
 v[0] = s1v[0] ^ s2v[0];
 v[1] = s1v[1] ^ s2v[1];
 return(GOOD);
}
/*****************************************************************************/
symone(s1,s2,v) 
char     s1[], s2[];
B32      v[];
{
B32 s1v[2];
int status;
 if(eval(s1,s1v) == OBAD) return(OBAD);
 v[0] = s1v[0];
 v[1] = s1v[1];
 return(GOOD);
}

