%{
/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <string.h>
#include <ctype.h>

#define NO_LEXYY
#include "pp.h"
#undef NO_LEXYY

#if defined(VMS) || defined(__MSDOS__)
#include "pp_tab.h"
#else
#include "pp.tab.h"
#endif

/* 	Macro to strip the quotes off a quoted string.  No checking is performed,
   but that's okay because we won't execute unless the token matches.			*/
#define UNQUOTE(o, i)	strncpy(o, &i[1], strlen(i) - 2)

extern int yychar;
extern YYSTYPE yylval;

YYSTYPE make_token PROTO((char *text, int type));
void get_input PROTO((char *buf, int *result, int max_size));

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size)	get_input(buf,&(result),max_size);
%}

%x expr
%x if_direct
%x if_body repeat_body
%x iif_body
%x kill_to_eol kill_to_eol_cmt
%x macro_args macro_value macro_body
%x psect_body

WS	[ \t]

%%
	static int dont_echo_ws = 0, dont_echo_nl = 0;
	static int expr_return = INITIAL, expr_return_comma = INITIAL;
	static int could_be_macro = 1;
	static int repeat_level = 0, if_level = 0;
	static int macro_val_comma = INITIAL, macro_val_nl = INITIAL;

{WS}*;[^\n]*			{ yylval = make_token(yytext, COMMENT); return COMMENT; }
{WS}*				{ if (!dont_echo_ws) { add_to_line(yytext); } }

"\n"				{ could_be_macro = 1; dont_echo_ws = 0; return EOL; }

("\f"|"/")			;		/* the assembler doesn't want to see the "/" in ADDF/SUC, of FF's */

".include"			{ dont_echo_ws = 1; return INCLUDE; }

".if"				{ dont_echo_ws = 1; if_level++; 
                                  macro_val_comma = macro_value; macro_val_nl = if_body;
				  expr_return = expr_return_comma = if_body; 
				  BEGIN(if_direct); return IF;
				}

".iif"				{ macro_val_comma = iif_body; macro_val_nl = INITIAL;
				  expr_return_comma = iif_body; expr_return = INITIAL;
				  BEGIN(if_direct); return IIF;
				}

<if_direct>("eq"|"equal")	{ yylval = make_token(NULL, EQ); BEGIN(expr); return EQ; }

<if_direct>("ne"|"not_equal")	{ yylval = make_token(NULL, NE); BEGIN(expr); return NE; }

<if_direct>("gt"|"greater")	{ yylval = make_token(NULL, GT); BEGIN(expr); return GT; }

<if_direct>("ge"|"greater_equal") { yylval = make_token(NULL, GE); BEGIN(expr); return GE; }

<if_direct>("lt"|"less_than")	{ yylval = make_token(NULL, LT); BEGIN(expr); return LT; }

<if_direct>("le"|"less_equal")	{ yylval = make_token(NULL, LE); BEGIN(expr); return LE; }

<if_direct>("df"|"defined")	{ yylval = make_token(NULL, DF); BEGIN(expr); return DF; }

<if_direct>("ndf"|"not_defined") { yylval = make_token(NULL, NDF); BEGIN(expr); return NDF; }

<if_direct>("b"|"blank")	{ yylval = make_token(NULL, B); BEGIN(macro_value); return B; }

<if_direct>("nb"|"not_blank")	{ yylval = make_token(NULL, NB); BEGIN(macro_value); return NB; }

<if_direct>.			;

<if_body>(".if_false"|".iff")	{ if (if_level == 1) {
                                    dont_echo_nl = 1;
				    yylval = make_token(line, IFF);
				    clear_line();
				    return IFF;
				  } else
				    add_to_line(yytext);
				}

<if_body>(".if_true"|".ift")	{ if (if_level == 1) {
                                    dont_echo_nl = 1;
				    yylval = make_token(line, IFT);
				    clear_line();
				    return IFT;
				  } else
				    add_to_line(yytext);
				}


<if_body>(".if_true_false"|".iftf") { if (if_level == 1) {
                                        dont_echo_nl = 1;
					yylval = make_token(line, IFTF);
					clear_line();
					return IFTF;
				      } else
					add_to_line(yytext);
				    }


<if_body>".if"			{ ++if_level; add_to_line(yytext); }

<if_body>".endc"		{ if (--if_level == 0) {
                                    dont_echo_nl = 0;
				    yylval = make_token(line, ENDC);
				    clear_line();
				    expr_return = expr_return_comma = INITIAL;
				    BEGIN(INITIAL);
				    return ENDC; 
				  } else
				    add_to_line(yytext);
				}

<if_body>{WS}*;[^\n]*		{ add_to_line(yytext); }

<if_body>.			{ add_to_line(yytext); }

<if_body>\n			{ if (dont_echo_nl) dont_echo_nl = 0; else add_to_line(yytext); }

<iif_body>[^\n]*		{ /* Must add NL to end of yytext.  strcat(yytext,"\n") overwrites something,
				     so let's do it the cheap (and slow) way. */
                                  add_to_line(yytext); add_to_line("\n");
				  yylval = make_token(line, ENDC);
				  clear_line(); BEGIN(INITIAL);
				  return ENDC;
				}

".repeat"			{ dont_echo_ws = 1; repeat_level++;
                                  expr_return = expr_return_comma = repeat_body;
				  BEGIN(expr); return REPEAT;
				}

<repeat_body>".repeat"		{ ++repeat_level; add_to_line(yytext); }

<repeat_body>{WS}*;[^\n]*	{ add_to_line(yytext); }

<repeat_body>.			{ add_to_line(yytext); }

<repeat_body>\n			{ add_to_line(yytext); }

<repeat_body>".endr"		{ if (--repeat_level == 0) { 
                                    yylval = make_token(line, ENDR);
				    clear_line();
				    expr_return = expr_return_comma = INITIAL;
				    BEGIN(INITIAL);
				    return ENDR;
				  } else
				    add_to_line(yytext);
				}

".macro"			{ dont_echo_ws = 1; BEGIN(macro_args); return MACRO; }

<macro_args>[a-zA-Z0-9_$.]+	{ yylval = make_token(yytext, SYMBOL); BEGIN(macro_args); return SYMBOL; }

<macro_args>"="			{ yylval = make_token(NULL, yytext[0]);
                                  macro_val_comma = macro_args; macro_val_nl = macro_body;
				  BEGIN(macro_value); return yytext[0];
				}

<macro_args>{WS}*;[^\n]*	{ yylval = make_token(yytext, COMMENT); return COMMENT; }

<macro_args>({WS}|"\f")*	;

<macro_args>(","|"?")		{ yylval = make_token(NULL, yytext[0]); return yytext[0]; }

<macro_args>\n			{ BEGIN(macro_body); return EOL; }

<macro_value>[^ ,;\t\f\n]*	{ yylval = make_token(yytext, ARG_VALUE); return ARG_VALUE; }

<macro_value>"<"[^>]*">"	{ yylval = make_token(yytext, STRING); yylval->type = ARG_VALUE; return ARG_VALUE; }

<macro_value>{WS}*;[^\n]*	{ yylval = make_token(yytext, COMMENT); return COMMENT; }

<macro_value>","		{ yylval = make_token(NULL, yytext[0]); BEGIN(macro_val_comma); return yytext[0]; }

<macro_value>"\n"		{ if (macro_val_nl == INITIAL)
                                    could_be_macro = 1;
				  BEGIN(macro_val_nl);
				  return EOL;
				}

<macro_body>.			{ add_to_line(yytext); }

<macro_body>\n			{ add_to_line(yytext); }

<macro_body>".endm"		{ yylval = make_token(line, ENDM); clear_line();
                                  expr_return = expr_return_comma = INITIAL;
				  BEGIN(INITIAL); return ENDM;
				}

".psect"			{ BEGIN(psect_body); return PSECT; }

<psect_body>[0-9]		{ yylval = cvtradix(yytext, 10, NUMBER); return NUMBER; }

<psect_body>[a-zA-Z0-9_$.]+	{ yylval = make_token(yytext, SYMBOL); return SYMBOL; }

<psect_body>","			{ return yytext[0]; }

<psect_body>{WS}*;[^\n]*	{ yylval = make_token(yytext, COMMENT); return COMMENT; }

<psect_body>{WS}*		;

<psect_body>"\n"		{ BEGIN(INITIAL); return EOL; }

(".save_psect"|".save")		{ BEGIN(kill_to_eol_cmt); return SAVE_PSECT; }

(".restore_psect"|".restore")	{ BEGIN(kill_to_eol_cmt); return RESTORE_PSECT; }

<kill_to_eol_cmt>{WS}*;[^\n]*	{ yylval = make_token(yytext, COMMENT); return COMMENT; }

<kill_to_eol_cmt>.		;

<kill_to_eol_cmt>\n		{ could_be_macro = 1; BEGIN(INITIAL); return EOL; }

{WS}*(".address"|".quad"|".long"){WS}*		{ expr_return = INITIAL; expr_return_comma = expr;
						  yylval = make_token(yytext, MEM_DIRECTIVE);
						  BEGIN(expr); return MEM_DIRECTIVE;
						}

".ascii"			{ return ASCII; }

".asciz"			{ return ASCIZ; }

".blkl"{WS}*			{ expr_return = expr_return_comma = INITIAL; add_to_line(yytext); BEGIN(expr); }

".blkq"{WS}*			{ expr_return = expr_return_comma = INITIAL; add_to_line(yytext); BEGIN(expr); }

".word"				{ errout(E_ERR, "yylex", ".word not supported.\n"); BEGIN(kill_to_eol); }

".byte"				{ errout(E_ERR, "yylex", ".byte not supported.\n"); BEGIN(kill_to_eol); }

<kill_to_eol>[^\n]*		;

<kill_to_eol>\n			{ could_be_macro = 1; clear_line(); BEGIN(INITIAL); return EOL; }

".align"{WS}*			{ expr_return_comma = expr; expr_return = INITIAL; add_to_line(yytext); BEGIN(expr); return ALIGN; }

","				{ expr_return = expr_return_comma = INITIAL; add_to_line(yytext); BEGIN(expr); }

","{WS}*"#"			{ expr_return = expr_return_comma = INITIAL; add_to_line(yytext); BEGIN(expr); }

[a-zA-Z0-9_$]+/({WS}*("="|"=="))	{ expr_return = expr_return_comma = INITIAL; BEGIN(expr); yylval = make_token(yytext, SYMBOL); return SYMBOL; }

"."/({WS}*("="|"=="))		{ expr_return = expr_return_comma = INITIAL; BEGIN(expr); yylval = make_token(NULL, DOT); return DOT; }

<expr>"^x"[0-9a-fA-F]+		{ yylval = cvtradix(&yytext[2], 16, NUMBER); return NUMBER; }

<expr>"^b"[0-1]+		{ yylval = cvtradix(&yytext[2], 2, NUMBER); return NUMBER; }

<expr>"^o"[0-7]+		{ yylval = cvtradix(&yytext[2], 8, NUMBER); return NUMBER; }

<expr>[0-9]+			{ yylval = cvtradix(yytext, 10, NUMBER); return NUMBER; }

<expr>"^c"			{ yylval = make_token(NULL, NOT); return NOT; }

<expr>("R"|"F")[0-9]([0-9])?	{ add_to_line(yytext); BEGIN(expr_return); }

<expr>"("			{ BEGIN(expr_return); return '('; }

<expr>","			{ BEGIN(expr_return_comma); yylval = make_token(NULL, ','); return ','; }

<expr>"\n"			{ could_be_macro = 1; BEGIN(expr_return); return EOL; }

<expr>[a-zA-Z0-9_$]+		{ yylval = make_token(yytext, SYMBOL); return SYMBOL; }

<expr>"."			{ yylval = make_token(NULL, DOT); return DOT; }

<expr>("+"|"-"|"&"|"!"|"\\"|"@"|"="|"=="|"<"|">"|"*")	{ yylval = make_token(NULL, yytext[0]); return yytext[0]; }

<expr>({WS}|"\f")*		;

<expr>{WS}*;[^\n]*		{ yylval = make_token(yytext, COMMENT); return COMMENT; }

\"([^\"]|\\\")*\"		{ yylval = make_token(yytext, STRING); return STRING; }

[a-zA-Z0-9_$]+/(":"|"::")	{ add_to_line(yytext); }

[a-zA-Z0-9_$.]+			{ if (could_be_macro) { 
				    could_be_macro = 0;
				    if (is_macro(yytext)) {
				      dont_echo_ws = 0;
				      macro_val_comma = macro_value;
				      macro_val_nl = INITIAL;
				      BEGIN(macro_value);
				      yylval = make_token(yytext, MACRO_CALL);
				      return MACRO_CALL;
				    }
				    /* fall through, this wasn't a macro, so add it to line */
				  }
				  add_to_line(yytext);
				}

.				{ add_to_line(yytext); }

<<EOF>>				{ if (cur_ifile->type == F_FILE) 
				    fclose(cur_ifile->f);   /* close the file we just finished */

				  if (--(cur_ifile->repeat) > 0) {
				    /* repeat count non-zero, push input back on stack */
				    cur_ifile->next = ifile;
				    ifile = cur_ifile;
				    ifile->f = (FILE *)ifile->name;
				    yy_delete_buffer(ifile->buff);
				    ifile->buff = NULL;
				    ifile->yylval = NULL;
				    ifile->yychar = YYEMPTY;
				  } else {
				    /* if this was a macro, pop macro arguments off stack. */
				    if (cur_ifile->type == F_MACRO) {
				      if (macstk == NULL) {
					errout(E_ERR, "yylex", "macro stack already NULL when popping.\n");
				      } else {
					MACRO_STK *junk = macstk;
					
					/* pop macro args off stack */
					macstk = macstk->next;
					free_macro_args(junk);
				      }
				    }
				    
				    /* deallocate storage associated with cur_ifile */
				    free_file(cur_ifile);
				  }
                                  if (!next_file())
                                    yyterminate();

                                  /* if there was a token pending, return it
                                     otherwise just keep going */
                                  if (yychar != YYEMPTY)
                                     return yychar;
                                }
%%

#ifdef __STDC__
YYSTYPE cvtradix(char *text, int radix, int type)
#else
YYSTYPE cvtradix(text, radix, type)
char *text;
int radix, type;
#endif
{
  YYSTYPE temp = ALLOC0(1, TOKEN_LIST);
  unsigned int32 *num = &temp->tok.q.v[0];
  unsigned int32 digit[2];
  int overflow = 0;

  temp->type = type;

  digit[1] = 0;
  while (*text) {
    overflow |= mul64(num, radix);

    if (isdigit(*text)) {
      digit[0] = *text - '0';
      add64(num, digit, num);
    } else {
      digit[0] = toupper(*text) - 'A' + 10;
      add64(num, digit, num);
    }
    text++;
  }

  if (overflow)
    errout(E_WARN, "cvtradix", "integer constant overflow truncating upper bits.\n");
  return temp;
}

#ifdef __STDC__
YYSTYPE make_token(char *text, int type)
#else
YYSTYPE make_token(text, type)
char *text;
int type;
#endif
{
  char *ctmp;
  YYSTYPE temp = ALLOC0(1, TOKEN_LIST);

  temp->type = type;
  switch (type) {
  case STRING:
    temp->tok.s = UNQUOTE(ALLOC_STR(strlen(text)-2), text);
    break;
  case SYMBOL:
    /* convert to lowercase first */
    for (ctmp = text; *ctmp; ++ctmp)
      *ctmp = tolower(*ctmp);
    /* fall through */
  case ARG_VALUE:
  case MACRO_CALL:
  case ENDM:
  case MEM_DIRECTIVE:
  case COMMENT:
  case LABEL:
  case IFF:
  case IFT:
  case IFTF:
  case ENDC:
  case ENDR:
    temp->tok.s = strcpy(ALLOC_STR(strlen(text)), text);
    break;
  default:
    /* an operator or DOT.  the type is the important part */
    break;
  }

  return temp;
}

#ifdef __STDC__
void add_to_line(char *text)
#else
void add_to_line(text)
char *text;
#endif
{
  if ((lptr + strlen(text) - line) > line_size) {
    int diff = lptr - line;

    line_size += 2048;
    line = (char *)realloc(line, line_size + 1);
    lptr = line + diff;
  }
    
  strcpy(lptr, text);
  lptr += strlen(text);

  return;
}

void clear_line()
{
  /* reset line */
  lptr = line;
  *lptr = '\0';

  return;
}

#ifdef __STDC__
void output_line(YYSTYPE comment)
#else
void output_line(comment)
YYSTYPE comment;
#endif
{
  if (comment == NULL)
    fprintf(yyout, "%s\n", line);
  else {
    fprintf(yyout, "%s%s\n", line, comment->tok.s);
    free_tok(comment);
  }

  clear_line();

  return;
}

void get_input(buf,result,max_size)
     char *buf;
     int *result;
     int max_size;
{
  char *sptr;
  int num;
  
  *buf = 0;		/* preset so end-of-file is detected... */
  if (cur_ifile->type == F_FILE) {
    fgets(buf,max_size,cur_ifile->f);
    cur_ifile->lineno++;
    *result = strlen(buf);
  } else 
    if (cur_ifile->type == F_STRING) {
      sptr = (char *)cur_ifile->f;
      if (sptr == NULL)
	*result = YY_NULL;
      else {
	num = strlen(sptr);
	if (num > max_size)
	  num = max_size;
	strncpy(buf,sptr,max_size);
	*result = num;
	if (strlen(sptr) > num) {
	  cur_ifile->f = (FILE *) (sptr + num);
	} else
	  cur_ifile->f = NULL;
      }
    } else 
      if (cur_ifile->type == F_MACRO) {
	static char *line_to_use = NULL;

	/* something left over from the last line?? */
	if (line_to_use != NULL) {
	  if ((num = strlen(line_to_use)) > max_size) {
	    strncpy(buf, line_to_use, max_size);
	    line_to_use += max_size;
	    *result = max_size;
	  } else {
	    strncpy(buf, line_to_use, num);
	    FREE(line_to_use);
	    line_to_use = NULL;
	    *result = num;
	  }
	  return;
	}

	sptr = (char *)cur_ifile->f;
	if (sptr == NULL) {
	  *result = YY_NULL;
	} else {
	  /* return macro body one line at a time, passing through the preprocessor first
	     NOTE: macro will always end in a "\n" because of parsing considerations */
	  char *ctmp, *tline;
	  int tlen;

	  if ((ctmp = strchr(sptr, '\n')) == NULL) {
	    /* well, this shouldn't happen, but it is harder for the lexer to prevent it,
	       so we'll handle it here.  If there is no NL at the end of the macro, just use
	       what's there. */
	    ctmp = sptr + strlen(sptr) - 1;
	  }
	  
	  /* make temporary copy of line */
	  tlen = ctmp - sptr + 1;
	  tline = strncpy(ALLOC_STR(tlen), sptr, tlen);
	  tline[tlen] = '\0';

	  /* increment here so that any errors received in prep_macro_line give the right 
	     line number */
	  cur_ifile->lineno++;

	  /* evaluate any %extract, %locate or %length functions */
	  prep_macro_line(tline, &line_to_use);
	  FREE(tline);

	  if ((num = strlen(line_to_use)) > max_size) {
	    strncpy(buf, line_to_use, max_size);
	    line_to_use += max_size;
	    *result = max_size;
	  } else {
	    strncpy(buf, line_to_use, num);
	    FREE(line_to_use);
	    line_to_use = NULL;
	    *result = num;
	  }

	  if (strlen(sptr) <= num)
	    cur_ifile->f = NULL;
	  else
	    cur_ifile->f = (FILE *)((char *)ctmp + 1);
	}
      }
  
  return;
}
