%{
/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include "pp.h"

extern FILE *yyin, *yyout;
int expr_contained_dot = 0;

YYSTYPE chain PROTO((YYSTYPE t1, YYSTYPE t2, YYSTYPE t3));
YYSTYPE create_zero_tok PROTO((void));

%}

%token NUMBER IF INCLUDE PLUS MINUS SHIFT EQUAL NOT OPEN_ANGLE CLOSE_ANGLE
%token COLON SYMBOL LABEL STRING EOL EOE DOT EQ GT LT DF NE GE LE NDF ENDC
%token ALIGN IFF IFT IFTF REPEAT ENDR COMMENT MEM_DIRECTIVE MACRO ENDM
%token MACRO_CALL ARG_VALUE ASCII ASCIZ SAVE_PSECT RESTORE_PSECT PSECT
%token B NB
%token IIF

%left '+' '-' '@' '&' '!' '\\' '*'
%right NOT UMINUS

%%
toplevel:
	sequence
    ;

sequence:
	/* empty */
    |	stuff
    |	sequence stuff
    ;

stuff:
	eol_w_cmt			{ output_line($1); }
    |	directive
    |   ','				{ add_to_line(","); }
    |   '('				{ add_to_line("("); }
    |	expression '('			{ enter_expression($1, '('); expr_contained_dot = 0; }
    |	expression ','			{ enter_expression($1, ','); expr_contained_dot = 0; }
    |	expression eol_w_cmt		{ enter_expression($1, EOL); expr_contained_dot = 0; output_line($2); }
    ;

directive:
	INCLUDE STRING eol_w_cmt				{ do_include($2); clear_line(); output_line($3); }
    |	ASCII STRING eol_w_cmt					{ do_ascii($2); output_line($3); }
    |	ASCIZ STRING eol_w_cmt					{ do_asciz($2); output_line($3); }
    |	IF condition if_value eol_w_cmt ifbody ENDC eol_w_cmt	{ do_if($2, $3, $5, $6); free_tok($4); expr_contained_dot = 0; clear_line(); output_line($7); }
    |	IIF condition if_value ',' ENDC	eol_w_cmt		{ do_if($2, $3, NULL, $5); expr_contained_dot = 0; clear_line(); output_line($6); }
    |	REPEAT expression eol_w_cmt ENDR eol_w_cmt		{ do_repeat($2, $4); free_tok($3); expr_contained_dot = 0; clear_line(); output_line($5); }
    |   SYMBOL '=' expression eol_w_cmt				{ define_symbol($1, $3); expr_contained_dot = 0; clear_line(); output_line($4); }
    |   DOT '=' expression eol_w_cmt				{ dot_equal($1,$3); expr_contained_dot = 0; clear_line(); output_line($4); }
    |	ALIGN exprseq eol_w_cmt					{ do_align($2); expr_contained_dot = 0; output_line($3); }
    |	MEM_DIRECTIVE exprseq eol_w_cmt				{ do_mem_directive($1, $2, $3); expr_contained_dot = 0; }
    |	MACRO SYMBOL macro_arglist eol_w_cmt ENDM eol_w_cmt	{ define_macro($2, $3, $5); free_tok($4); expr_contained_dot = 0; clear_line(); output_line($6); }
    |	MACRO_CALL macro_call_args eol_w_cmt			{ expand_macro($1, $2); expr_contained_dot = 0; output_line($3); }
    |	PSECT SYMBOL psect_args eol_w_cmt			{ do_psect($2, $3); expr_contained_dot = 0; output_line($4); }
    |	SAVE_PSECT eol_w_cmt					{ do_save_psect(); output_line($2); }
    |	RESTORE_PSECT eol_w_cmt					{ do_rest_psect(); output_line($2); }
    ;

psect_args:
	/* empty */			{ $$ = NULL; }
    |	NUMBER				{ $$ = $1; }
    |	SYMBOL				{ $$ = $1; }
    |	psect_args ',' NUMBER		{ if ($1 == NULL) $$ = $3; else $$ = chain($1, $3, NULL); }
    |	psect_args ',' SYMBOL		{  if ($1 == NULL) $$ = $3; else $$ = chain($1, $3, NULL); }
    ;

eol_w_cmt:
	EOL				{ $$ = NULL; }
    |	COMMENT EOL			{ $$ = $1; }
    ;

macro_call_args:
        opt_value			{ $$ = $1; }
    |   macro_call_args ',' opt_value   { if ($1 == NULL) 
	                                    $$ = chain($2, $3, NULL);
                                          else 	
					    $$ = chain ($1, $2, $3);
					}
    ;

opt_value:
	/* empty*/			{ $$ = NULL; }
    |	ARG_VALUE			{ $$ = $1; }
    ;

macro_arglist:
	macro_arg			{ $$ = $1; }
    |	macro_arglist ',' macro_arg	{ $$ = chain($1, $2, $3); }

macro_arg:
	/* empty */			{ $$ = NULL; }
    |	SYMBOL arg_default		{ $$ = chain($1, $2, NULL); }
    |	'?' SYMBOL			{ $$ = chain($2, $1, NULL); }
    ;

arg_default:
	/* empty */			{ $$ = NULL; }
    |	'=' ARG_VALUE			{ $$ = chain($1, $2, NULL); }
    ;

if_value:
	/* empty */			{ $$ = NULL; }
    |	expression			{ $$ = $1; }
    |	ARG_VALUE			{ $$ = $1; }
    ;

ifbody:
	/* empty */			{ $$ = NULL; }
    |	ifsubdirect			{ $$ = $1; }
    |	ifbody ifsubdirect		{ $$ = chain($1, $2, NULL); }
    ;

ifsubdirect:
	IFF				{ $$ = $1; }
    |	IFT				{ $$ = $1; }
    |	IFTF				{ $$ = $1; }
    ;
	
condition:
	EQ				{ $$ = $1; }
    |	GT				{ $$ = $1; }
    |	LT				{ $$ = $1; }
    |	DF				{ $$ = $1; }
    |	NE				{ $$ = $1; }
    |	GE				{ $$ = $1; }
    |	LE				{ $$ = $1; }
    |	NDF				{ $$ = $1; }
    |	B				{ $$ = $1; }
    |	NB				{ $$ = $1; }
    ;

exprseq:
	expression			{ $$ = $1; }
    |	exprseq ',' expression		{ $$ = chain($1, $2, $3); }
    ;

expression:
	element				{ $$ = $1; }
    |	expression '+' expression	{ $$ = chain($1, $3, $2); }
    |	expression '-' expression	{ $$ = chain($1, $3, $2); }
    |	expression '*' expression	{ $$ = chain($1, $3, $2); }
    |	expression '@' expression	{ $$ = chain($1, $3, $2); }
    |	expression '&' expression	{ $$ = chain($1, $3, $2); }
    |	expression '!' expression	{ $$ = chain($1, $3, $2); }
    |	expression '\\' expression	{ $$ = chain($1, $3, $2); }
    |	NOT expression			{ $$ = chain($2, $1, NULL); }
    |	'-' expression %prec UMINUS	{ $$ = chain(create_zero_tok(), $2, $1); }
    |	'<' expression '>'		{ $$ = $2; }
    ;

element:
	NUMBER					{ $$ = $1; }
    |	SYMBOL					{ $$ = $1; }
    |	DOT					{ $$ = $1; expr_contained_dot = 1; }
    ;
%%

#ifdef __STDC__
yyerror(char *s)
#else
yyerror(s)
char *s;
#endif
{
  errout(E_ERR, "bison", "%s\n", s);

  return;
}

#ifdef __STDC__
YYSTYPE chain(YYSTYPE t1, YYSTYPE t2, YYSTYPE t3)
#else
YYSTYPE chain(t1, t2, t3)
YYSTYPE t1;
YYSTYPE t2;
YYSTYPE t3;
#endif
{
  YYSTYPE ttmp;

  if (t1) {
    for (ttmp = t1; ttmp->next; ttmp = ttmp->next)
      ;
    ttmp->next = t2;
  }
  if (t2) {
    for (ttmp = t2; ttmp->next; ttmp = ttmp->next)
      ;
    ttmp->next = t3;
  }

  return t1;
}

YYSTYPE create_zero_tok()
{
  YYSTYPE temp = ALLOC0(1, TOKEN_LIST);

  temp->type = NUMBER;

  return temp;
}
