/*
	Copyright (c) 1992
	Digital Equipment Corporation, Maynard, Massachusetts 01754

	The HAL Alpha Assembler is free software; you can redistribute 
	it and/or modify it with the sole restriction that this copyright 
	notice be attached.

	The information in this software is subject to change without notice 
	and should not be construed as a commitment by Digital Equipment
	Corporation.

	The author disclaims all warranties with regard to this software,
	including all implied warranties of merchantability and fitness, in
	no event shall he be liable for any special, indirect or consequential
	damages or any damages whatsoever resulting from loss of use, data, 
	or profits, whether in an action of contract, negligence or other
	tortious action, arising out of or in connection with the use or
	performance of this software.

*/
#include <stdio.h>
#include <string.h>

#include "pp.h"
#if defined(VMS) || defined(__MSDOS__)
#include "pp_tab.h"
#else
#include "pp.tab.h"
#endif

extern FILE *yyout;
extern int expr_contained_dot;

#ifdef __STDC__
TOKEN_LIST *pop_token(TOKEN_LIST **head)
#else
TOKEN_LIST *pop_token(head)
TOKEN_LIST **head;
#endif
{
  TOKEN_LIST *temp = *head;

  if (*head != NULL)
    *head = (*head)->next;

  return temp;
}

#ifdef __STDC__
void push_token(TOKEN_LIST *temp, TOKEN_LIST **stack)
#else
void push_token(temp, stack)
TOKEN_LIST *temp, **stack;
#endif
{
  if (temp != NULL) {
    temp->next = *stack;
    *stack = temp;
  }

  return;
}

#ifdef __STDC__
TOKEN_LIST *add_assign(TOKEN_LIST *name, TOKEN_LIST *op1, TOKEN_LIST *op2, TOKEN_LIST *oper)
#else
TOKEN_LIST *add_assign(name, op1, op2, oper)
TOKEN_LIST *name, *op1, *op2, *oper;
#endif
{
  char junk[50],*ctmp;
  ASSIGN *temp = ALLOC0(1, ASSIGN);
  TOKEN_LIST *ttemp = NULL;

  if (op1)
    op1->next = op2;
  if (op2)
    op2->next = oper;
  if (oper)
    oper->next = NULL;
  temp->operand1 = op1;
  temp->operand2 = op2;
  temp->operator = oper;

  if ((name != NULL) && (name->type == DOT)) {
    temp->name = make_token(NULL, DOT);
  } else if (name != NULL) {
    SYMDEF *stemp;

    if (!avl_lookup(symtree, name->tok.s, &stemp)) {
      stemp = ALLOC0(1, SYMDEF);
      stemp->incarnation = 0;
      avl_insert(symtree, name->tok.s, stemp);
    }

    if (stemp->incarnation == S_LABEL)
      junk[0] = '\0';
    else
      sprintf(junk, "%%%d", stemp->incarnation++);

    ctmp = strcpy(ALLOC_STR(strlen(junk) + strlen(name->tok.s)), name->tok.s);
    strcat(ctmp, junk);
    temp->name = make_token(ctmp, SYMBOL);
  } else {
    SYMDEF *stemp;
    /* create a new temporary variable */
    if (!avl_lookup(symtree, "%", &stemp)) {
      char *ctmp = ALLOC_STR(1);

      strcpy(ctmp, "%");
      stemp = ALLOC0(1, SYMDEF);
      stemp->incarnation = 0;
      avl_insert(symtree, ctmp, stemp);
    }

    sprintf(junk, "%%%d", stemp->incarnation++);
    temp->name = make_token(strcpy(ALLOC_STR(strlen(junk)), junk), SYMBOL);
  }

  temp->next = defs;
  defs = temp;

  return temp->name;
}

#ifdef __STDC__
void modify_symbol(TOKEN_LIST *sym, int i)
#else
void modify_symbol(sym, i)
TOKEN_LIST *sym;
int i;
#endif
{
  char *temp, junk[20];

  /* create suffix to add to symbol name */
  sprintf(junk, "%%%d", i);

  /* save old value */
  temp = sym->tok.s; 

  /* create new name */
  sym->tok.s = ALLOC_STR(strlen(temp)+strlen(junk));
  strcpy(sym->tok.s, temp);
  strcat(sym->tok.s, junk);

  return;
}

#ifdef __STDC__
void print_token_stack(TOKEN_LIST *temp)
#else
void print_token_stack(temp)
TOKEN_LIST *temp;
#endif
{
  
  while (temp != NULL) {
    switch (temp->type) {
    case NUMBER:
      fprintf(stderr,"%d ", temp->tok.q.v[0]);
      break;
    case SYMBOL:
      fprintf(stderr,"%s ", temp->tok.s);
      break;
    case NOT:
      fprintf(stderr,"^c ");
      break;
    case DOT:
      fprintf(stderr,". ");
      break;
    case '+':
    case '-':
    case '*':
    case '@':
    case '&':
    case '!':
    case '\\':
    case '=':
    case '<':
    case '>':
      fprintf(stderr,"%c ", temp->type);
      break;
    }
    temp = temp->next;
  }
  fprintf(stderr, "\n");

  return;
}

#ifdef __STDC__
void eval_expr(TOKEN_LIST *temp, TOKEN_LIST **return_stack)
#else
void eval_expr(temp, return_stack)
TOKEN_LIST *temp, **return_stack;
#endif
{
  TOKEN_LIST *stack = *return_stack;

  while (temp != NULL) {
    SYMDEF *stemp;
    TOKEN_LIST *op1, *op2;
    TOKEN_LIST *ntemp = temp->next;   /* save next pointer */

#ifdef DEBUG
    print_token_stack(stack);
#endif

    switch (temp->type) {
    case SYMBOL:
      /* what incarnation are we talking about?? */
      if (avl_lookup(symtree, temp->tok.s, &stemp)) {
	if (stemp->incarnation > 0)
	  modify_symbol(temp, stemp->incarnation - 1);
      } else {
	/* if the first use of the symbol is as an operand, it had better 
	   be a label, later on.  Record that info here. */
	stemp = ALLOC0(1, SYMDEF);
	stemp->incarnation = S_LABEL;
	avl_insert(symtree, temp->tok.s, stemp);
      }
      /* fall through to push operand on stack */
    case DOT:
    case NUMBER:
      /* push operand on stack */
      push_token(temp, &stack);
      break;
    case '+':
    case '-':
    case '*':
    case '@':
    case '&':
    case '!':
    case '\\':
      op2 = pop_token(&stack);
      op1 = pop_token(&stack);
      
      if ((op1 == NULL) || (op2 == NULL)) {
	errout (E_ERR, "eval_expr", "Illegal number of operands for two operand operator\n");
	break;
      }
      
      if ((op1->type == NUMBER) && (op2->type == NUMBER)) {
	/* both operands are constants, let's evaluate it */
	switch (temp->type) {
	case '+':
	  op1 = addq(op1, op2);
	  break;
	case '-':
	  op1 = subq(op1, op2);
	  break;
	case '*':
	  op1 = mulq(op1, op2);
	  break;
	case '@':
	  op1 = shiftq(op1, op2);
	  break;
	case '&':
	  op1 = andq(op1, op2);
	  break;
	case '!':
	  op1 = orq(op1, op2);
	  break;
	case '\\':
	  op1 = xorq(op1, op2);
	  break;
	default:
	  errout(E_ERR, "eval_expr", "this couldn't have happened.\n");
	  break;
	}
	/* put result back on stack */
	push_token(op1, &stack);
      } else {
	/* one or both operands are symbols, create an assignment to a temp
	   variable, and push that symbol on the stack */
	push_token(add_assign(NULL, op1, op2, temp), &stack);
      }
      
      break;
    case NOT:
      {
	TOKEN_LIST *op1 = pop_token(&stack);

	if (op1 == NULL) {
	  errout (E_ERR, "eval_expr", "no operands for an operator\n");
	  break;
	}

	if (op1->type == NUMBER) {
	  /* operand is a constant, let's evaluate it */
	  switch (temp->type) {
	  case NOT:
	    op1 = notq(op1);
	    break;
	  default:
	    errout(E_ERR, "eval_expr", "this couldn't have happened.\n");
	    break;
	  }
	  /* put result back on stack */
	  push_token(op1, &stack);
	} else {
	  /* operand is a symbol, create an assignment to a temp
	     variable and that symbol on the stack */
	  push_token(add_assign(NULL, op1, NULL, temp), &stack);
	}

	break;
      }
    default:
      break;
    }
    temp = ntemp;
  }  

  *return_stack = stack;

  return;
}

#ifdef __STDC__
void define_symbol(TOKEN_LIST *name, TOKEN_LIST *exp)
#else
void define_symbol(name, exp)
TOKEN_LIST *name, *exp;
#endif
{
  TOKEN_LIST *stack = NULL, *temp = exp;

#ifdef DEBUG
  /* prints the symbol definition */
  fprintf(stderr,"symbol definition: %s\n", name->tok.s);
  print_token_stack(temp);
#endif

  eval_expr(temp, &stack);

  if ((stack == NULL) || (stack->next != NULL)) {
    errout(E_ERR, "define_symbol", "incorrect number of elements left on evaluation stack.\n");
    return;
  }

  /* okay, at this point we have one element on the stack, either DOT, a NUMBER, or a 
     SYMBOL.  If it is a NUMBER or DOT, just output the appropriate assignment.  If it is
     a temporary symbol (which we created, i.e. starts with a "%"), then alter the previous
     assignment to just assign directly to our variable. */
  if ((stack->type == SYMBOL) && (stack->tok.s[0] == '%')) {
    SYMDEF *stemp;
    ASSIGN *atemp;

    if (strcmp(defs->name->tok.s, stack->tok.s) != 0) {
      errout(E_ERR, "define_symbol", "last temp assignment does not match last evaluation element\n");
      defs = NULL;
      return;
    }
    atemp = defs;
    defs = defs->next;
    add_assign(name, atemp->operand1, atemp->operand2, atemp->operator);

    /* update temporary counter */
    if (avl_lookup(symtree, "%", &stemp)) {
      stemp->incarnation--;
    } else
      errout(E_ERR, "define_symbol", "temporary variable entry does not exists\n");
  } else 
    add_assign(name, stack, NULL, NULL);

  output_assignments(defs);
  defs = NULL;

  return;
}

#ifdef __STDC__
void dot_equal(TOKEN_LIST *name, TOKEN_LIST *exp)
#else
void dot_equal(name, exp)
TOKEN_LIST *name, *exp;
#endif
{
  TOKEN_LIST *stack = NULL, *temp = exp;
  ASSIGN *old_defs = defs;

#ifdef DEBUG
  /* prints the symbol definition */
  fprintf(stderr,"dot expression:\n");
  print_token_stack(temp);
#endif

  eval_expr(temp, &stack);

  if ((stack == NULL) || (stack->next != NULL)) {
    errout(E_ERR, "dot_equal", "incorrect number of elements left on evaluation stack.\n");
    return;
  }

  /* okay, at this point we have one element on the stack, either DOT, a NUMBER, or a 
     SYMBOL.  If it is a NUMBER or DOT, just output the appropriate assignment.  If it is
     a temporary symbol (which we created, i.e. starts with a "%"), then alter the previous
     assignment to just assign directly to our variable. */
  if ((stack->type == SYMBOL) && (stack->tok.s[0] == '%')) {
    SYMDEF *stemp;
    ASSIGN *atemp;

    if (strcmp(defs->name->tok.s, stack->tok.s) != 0) {
      errout(E_ERR, "dot_equal", "last temp assignment does not match last evalulation element\n");
      defs = NULL;
      return;
    }
    atemp = defs;
    defs = defs->next;
    add_assign(name, atemp->operand1, atemp->operand2, atemp->operator);

    /* update temporary counter */
    if (avl_lookup(symtree, "%", &stemp)) {
      stemp->incarnation--;
    } else
      errout(E_ERR, "dot_equal", "temporary variable entry does not exists\n");
  } else 
    add_assign(name, stack, NULL, NULL);

  output_assignments(defs);
  defs = NULL;

  return;
}

#ifdef __STDC__
void enter_expression(TOKEN_LIST *exp, int which)
#else
void enter_expression(exp, which)
TOKEN_LIST *exp;
int which;
#endif
{
  TOKEN_LIST *stack = NULL, *temp = exp;
  char str[2];

#ifdef DEBUG
  /* prints the expression */
  fprintf(stderr,"inline expression:\n");
  print_token_stack(temp);
#endif

  if (expr_contained_dot && dot_to_next_inst) {
    TOKEN_LIST *four = cvtradix("4", 10, NUMBER);
    TOKEN_LIST *plus = make_token(NULL, '+');
    TOKEN_LIST *ttmp;

    for (ttmp = temp; ttmp->next != NULL; ttmp = ttmp->next)
      ;

    ttmp->next = four;
    four->next = plus;
  }

  eval_expr(temp, &stack);

  output_assignments(defs);
  defs = NULL;

  if ((stack == NULL) || (stack->next != NULL)) {
    errout(E_ERR, "enter_expression", "incorrect number of elements left on evaluation stack.\n");
    return;
  }

  add_to_line(tok2str(stack));

  /* EOL handled differently now that we pass comments */
  if (which != EOL) {
    str[0] = (which == EOL) ? '\n' : which;
    str[1] = '\0';
    add_to_line(str);
  }
#if 0
  fprintf(yyout, "%s%c", tok2str(stack), (which == EOL) ? '\n' : which);
#endif

  return;
}

#ifdef __STDC__
void free_assign(ASSIGN *defs)
#else
void free_assign(defs)
ASSIGN *defs;
#endif
{
  free_tok(defs->name);
  if (defs->operand1 != NULL)
    if ((defs->operand1->type != SYMBOL) || (strchr(defs->operand1->tok.s, '%') == NULL))
      free_tok(defs->operand1);
  if (defs->operand2 != NULL)
    if ((defs->operand2->type != SYMBOL) || (strchr(defs->operand2->tok.s, '%') == NULL))
      free_tok(defs->operand2);
  if (defs->operator != NULL)
    if ((defs->operator->type != SYMBOL) || (strchr(defs->operator->tok.s, '%') == NULL))
      free_tok(defs->operator);
  FREE(defs);

  return;
}

#ifdef __STDC__
void output_assignments(ASSIGN *defs)
#else
void output_assignments(defs)
ASSIGN *defs;
#endif
{

  if (defs == NULL) {
    return;
  }

  output_assignments(defs->next);

  /* add assignment to symbol table for later use (if not a DOT assignment) */
  if (defs->name->type != DOT) {
    if (!avl_lookup(symtab, defs->name->tok.s, NULL)) {
      avl_insert(symtab, defs->name->tok.s, defs);
    } else
      /* already existed, this cannot happen because we create a new incarnation for
	 each assignment */
      errout(E_ERR, "output_assignments", "second assignment to same intermediate variable.\n");
  }

  /* send assignment to output file */
  fprintf (yyout, "%s = %s %s %s\n", tok2str(defs->name), tok2str(defs->operand1),
	   tok2str(defs->operator), tok2str(defs->operand2));

  if (defs->name->type == DOT)
    /* need to free this assignment structure if it is a ". = ..." */
    free_assign(defs);

  return;
}

#ifdef __STDC__
TOKEN_LIST *copy_expr (TOKEN_LIST *expr)
#else
TOKEN_LIST *copy_expr (expr)
TOKEN_LIST *expr;
#endif
{
  TOKEN_LIST *texpr = NULL;

  if (expr != NULL) {
    texpr = ALLOC0(1, TOKEN_LIST);
    memcpy(texpr, expr, sizeof(TOKEN_LIST));
    texpr->next = copy_expr(expr->next);
  }

  return texpr;
}

#ifdef __STDC__
TOKEN_LIST *eval_const_expr (TOKEN_LIST *expr)
#else
TOKEN_LIST *eval_const_expr (expr)
TOKEN_LIST *expr;
#endif
{
  TOKEN_LIST *stack = NULL;

  while (expr != NULL) {
    ASSIGN *atemp;
    SYMDEF *stemp;
    TOKEN_LIST *op1, *op2;
    TOKEN_LIST *ntemp = expr->next;   /* save next pointer */

#ifdef DEBUG
    print_token_stack(stack);
#endif

    switch (expr->type) {
    case SYMBOL:
      /* what incarnation are we talking about?? */
      if (strchr(expr->tok.s, '%') == NULL) {
	if (avl_lookup(symtree, expr->tok.s, &stemp)) {
	  if (stemp->incarnation > 0)
	    modify_symbol(expr, stemp->incarnation - 1);
	  else {
	    errout(E_ERR, "eval_const_expr", "symbol %s is a label (or was used prior to assignment).\n", expr->tok.s);
	    return NULL;
	  }
	} else {
#if 0    /* can't convert to 0 because of .if defined */
	  TOKEN_LIST *temp = cvtradix("0", 10, NUMBER);
	  /* symbol has no value */
	  
	  errout(E_ERR, "eval_const_expr", "symbol %s does not exist, using 0.\n", expr->tok.s);
	  temp->next = expr->next;
	  expr = temp;
#endif
	  return NULL;
	}
      }

      if (!avl_lookup(symtab, expr->tok.s, &atemp)) {
#if 0    /* can't convert to 0 because of .if defined */
	TOKEN_LIST *temp = cvtradix("0", 10, NUMBER);
	/* symbol has no value */
	
	errout(E_ERR, "eval_const_expr", "symbol %s does not exist, using 0.\n", expr->tok.s);
	temp->next = expr->next;
	expr = temp;
#endif
	return NULL;
      } else {
	TOKEN_LIST *texpr = copy_expr(atemp->operand1);
	expr = eval_const_expr(texpr);
	if (expr == NULL)
	  return NULL;
      }
      /* fall through to push operand on stack */
    case DOT:
      if (expr->type == DOT) {
	/* cannot include dot anywhere in a compile time constant expression */
	errout(E_ERR, "eval_const_expr", "dot is illegal in a constant expression.\n");
	return NULL;
      }
      /* fall through to push operand on stack */
    case NUMBER:
      /* push operand on stack */
      push_token(expr, &stack);
      break;
    case '+':
    case '-':
    case '*':
    case '@':
    case '&':
    case '!':
    case '\\':
      op2 = pop_token(&stack);
      op1 = pop_token(&stack);
      
      if ((op1 == NULL) || (op2 == NULL)) {
	errout (E_ERR, "eval_const_expr", "Illegal number of operands for two operand operator\n");
	break;
      }
      
      if ((op1->type == NUMBER) && (op2->type == NUMBER)) {
	/* both operands are constants, let's evaluate it */
	switch (expr->type) {
	case '+':
	  op1 = addq(op1, op2);
	  break;
	case '-':
	  op1 = subq(op1, op2);
	  break;
	case '*':
	  op1 = mulq(op1, op2);
	  break;
	case '@':
	  op1 = shiftq(op1, op2);
	  break;
	case '&':
	  op1 = andq(op1, op2);
	  break;
	case '!':
	  op1 = orq(op1, op2);
	  break;
	case '\\':
	  op1 = xorq(op1, op2);
	  break;
	default:
	  errout(E_ERR, "eval_const_expr", "this couldn't have happened.\n");
	  break;
	}
	/* put result back on stack */
	push_token(op1, &stack);
      } else {
	/* there better never be a symbol here */
	errout(E_ERR, "eval_const_expr", "symbol present where none can be.\n");
      }
      
      break;
    case NOT:
      if (op1 == NULL) {
	errout (E_ERR, "eval_const_expr", "no operands for an operator\n");
	break;
      }
      
      if (op1->type == NUMBER) {
	/* operand is a constant, let's evaluate it */
	switch (expr->type) {
	case NOT:
	  op1 = notq(op1);
	  break;
	default:
	  errout(E_ERR, "eval_const_expr", "this couldn't have happened.\n");
	  break;
	}
	/* put result back on stack */
	push_token(op1, &stack);
      } else {
	/* there better never be a symbol here */
	errout(E_ERR, "eval_const_expr", "symbol present where none can be.\n");
      }
      
      break;
    default:
      break;
    }
    expr = ntemp;
  }  

  return stack;
}
